@@ -43,11 +43,8 @@ Q_DECLARE_METATYPE(FileChooserPortal::Filter)
 Q_DECLARE_METATYPE(FileChooserPortal::Filters)
 Q_DECLARE_METATYPE(FileChooserPortal::FilterList)
 Q_DECLARE_METATYPE(FileChooserPortal::FilterListList)
-// used for options - choices
-Q_DECLARE_METATYPE(FileChooserPortal::Choice)
-Q_DECLARE_METATYPE(FileChooserPortal::Choices)
-Q_DECLARE_METATYPE(FileChooserPortal::Option)
-Q_DECLARE_METATYPE(FileChooserPortal::OptionList)
+
+using namespace Qt::StringLiterals;
 
 QDBusArgument &operator<<(QDBusArgument &arg, const FileChooserPortal::Filter &filter)
 {
@@ -91,54 +88,10 @@ const QDBusArgument &operator>>(const QDBusArgument &arg, FileChooserPortal::Fil
     return arg;
 }
 
-QDBusArgument &operator<<(QDBusArgument &arg, const FileChooserPortal::Choice &choice)
-{
-    arg.beginStructure();
-    arg << choice.id << choice.value;
-    arg.endStructure();
-    return arg;
-}
-
-const QDBusArgument &operator>>(const QDBusArgument &arg, FileChooserPortal::Choice &choice)
-{
-    QString id;
-    QString value;
-    arg.beginStructure();
-    arg >> id >> value;
-    choice.id = id;
-    choice.value = value;
-    arg.endStructure();
-    return arg;
-}
-
-QDBusArgument &operator<<(QDBusArgument &arg, const FileChooserPortal::Option &option)
-{
-    arg.beginStructure();
-    arg << option.id << option.label << option.choices << option.initialChoiceId;
-    arg.endStructure();
-    return arg;
-}
-
-const QDBusArgument &operator>>(const QDBusArgument &arg, FileChooserPortal::Option &option)
-{
-    QString id;
-    QString label;
-    FileChooserPortal::Choices choices;
-    QString initialChoiceId;
-    arg.beginStructure();
-    arg >> id >> label >> choices >> initialChoiceId;
-    option.id = id;
-    option.label = label;
-    option.choices = choices;
-    option.initialChoiceId = initialChoiceId;
-    arg.endStructure();
-    return arg;
-}
-
 static bool isKIOFuseAvailable()
 {
-    static bool available =
-        QDBusConnection::sessionBus().interface() && QDBusConnection::sessionBus().interface()->activatableServiceNames().value().contains("org.kde.KIOFuse");
+    static bool available = QDBusConnection::sessionBus().interface()
+        && QDBusConnection::sessionBus().interface()->activatableServiceNames().value().contains("org.kde.KIOFuse"_L1);
     return available;
 }
 
@@ -156,7 +109,7 @@ static QString decodeFileName(const QByteArray &name)
 FileDialog::FileDialog(QDialog *parent, Qt::WindowFlags flags)
     : QDialog(parent, flags)
     , m_fileWidget(new KFileWidget(QUrl(), this))
-    , m_configGroup(KSharedConfig::openConfig()->group("FileDialogSize"))
+    , m_configGroup(KSharedConfig::openConfig()->group(QStringLiteral("FileDialogSize")))
 {
     setLayout(new QVBoxLayout);
     layout()->setContentsMargins({});
@@ -300,9 +253,15 @@ static QUrl kioUrlFromSandboxPath(const QString &path, Entity entity)
                                                            QDBusConnection::sessionBus());
 
     if (mountPoint.isEmpty()) {
-        mountPoint = QString::fromUtf8(documents_iface.GetMountPoint());
-        if (!path.startsWith(mountPoint)) {
-            return QUrl::fromLocalFile(path);
+        auto reply = documents_iface.GetMountPoint();
+        reply.waitForFinished();
+        if (reply.isError()) {
+            qCWarning(XdgDesktopPortalKdeFileChooser) << "Failed to get documents mount point:" << reply.error();
+        } else {
+            mountPoint = QString::fromUtf8(reply);
+            if (!path.startsWith(mountPoint)) {
+                return QUrl::fromLocalFile(path);
+            }
         }
     }
 
@@ -320,12 +279,18 @@ static QUrl kioUrlFromSandboxPath(const QString &path, Entity entity)
     }
 
     if (!isKIOFuseAvailable()) {
-        return QUrl::fromLocalFile(localFilePath);
+        return QUrl::fromLocalFile(QString::fromLocal8Bit(localFilePath));
     }
 
     OrgKdeKIOFuseVFSInterface fuse_iface(QStringLiteral("org.kde.KIOFuse"), QStringLiteral("/org/kde/KIOFuse"), QDBusConnection::sessionBus());
 
-    QString remoteFilePath = fuse_iface.remoteUrl(localFilePath);
+    auto reply = fuse_iface.remoteUrl(QString::fromLocal8Bit(localFilePath));
+    reply.waitForFinished();
+    if (reply.isError()) {
+        qCWarning(XdgDesktopPortalKdeFileChooser) << "Failed to get fuse remote URL:" << reply.error();
+        return QUrl::fromLocalFile(path);
+    }
+    QString remoteFilePath = reply.value();
     if (remoteFilePath.isEmpty()) {
         return QUrl::fromLocalFile(path);
     }
@@ -336,12 +301,14 @@ static QUrl kioUrlFromSandboxPath(const QString &path, Entity entity)
     return url;
 };
 
-uint FileChooserPortal::OpenFile(const QDBusObjectPath &handle,
+void FileChooserPortal::OpenFile(const QDBusObjectPath &handle,
                                  const QString &app_id,
                                  const QString &parent_window,
                                  const QString &title,
                                  const QVariantMap &options,
-                                 QVariantMap &results)
+                                 const QDBusMessage &message,
+                                 uint &replyResponse,
+                                 QVariantMap &replyResults)
 {
     Q_UNUSED(app_id);
 
@@ -377,7 +344,9 @@ uint FileChooserPortal::OpenFile(const QDBusObjectPath &handle,
     const auto [filters, currentFilter] = ExtractFilters(options);
 
     if (isMobile()) {
-        if (!m_mobileFileDialog) {
+        if (m_mobileFileDialog) {
+            m_mobileFileDialog->reset(true);
+        } else {
             qCDebug(XdgDesktopPortalKdeFileChooser) << "Creating file dialog";
             m_mobileFileDialog = new MobileFileDialog(this);
         }
@@ -409,11 +378,11 @@ uint FileChooserPortal::OpenFile(const QDBusObjectPath &handle,
             m_mobileFileDialog->setMimeTypeFilters(filters.first().mimePatterns());
         }
 
-        uint retCode = m_mobileFileDialog->exec();
+        replyResponse = m_mobileFileDialog->exec();
 
-        results.insert(QStringLiteral("uris"), fuseRedirect(m_mobileFileDialog->results()));
+        replyResults.insert(QStringLiteral("uris"), fuseRedirect(m_mobileFileDialog->results()));
 
-        return retCode;
+        return;
     }
 
     const QUrl translatedCurrentFolderUrl = kioUrlFromSandboxPath(currentFolder, Entity::Folder);
@@ -421,56 +390,57 @@ uint FileChooserPortal::OpenFile(const QDBusObjectPath &handle,
     // Use QFileDialog for most directory requests to utilize
     // plasma-integration's KDirSelectDialog
     if (directory && !options.contains(QStringLiteral("choices"))) {
-        QFileDialog dirDialog;
-        dirDialog.setWindowTitle(title);
-        dirDialog.setModal(modalDialog);
-        dirDialog.setFileMode(QFileDialog::Directory);
-        dirDialog.setOptions(QFileDialog::ShowDirsOnly);
+        auto dirDialog = new QFileDialog;
+        dirDialog->setWindowTitle(title);
+        dirDialog->setFileMode(QFileDialog::Directory);
+        dirDialog->setOptions(QFileDialog::ShowDirsOnly);
         if (!isKIOFuseAvailable()) {
-            dirDialog.setSupportedSchemes(QStringList{QStringLiteral("file")});
+            dirDialog->setSupportedSchemes(QStringList{QStringLiteral("file")});
         }
         if (!acceptLabel.isEmpty()) {
-            dirDialog.setLabelText(QFileDialog::Accept, acceptLabel);
+            dirDialog->setLabelText(QFileDialog::Accept, acceptLabel);
         }
         if (!translatedCurrentFolderUrl.isEmpty()) {
-            dirDialog.setDirectoryUrl(translatedCurrentFolderUrl);
+            dirDialog->setDirectoryUrl(translatedCurrentFolderUrl);
         }
 
-        dirDialog.winId(); // Trigger window creation
-        Utils::setParentWindow(&dirDialog, parent_window);
-        Request::makeClosableDialogRequest(handle, &dirDialog);
+        dirDialog->winId(); // Trigger window creation
+        Utils::setParentWindow(dirDialog, parent_window);
+        Request::makeClosableDialogRequest(handle, dirDialog);
+        dirDialog->setWindowModality(modalDialog ? Qt::WindowModal : Qt::NonModal);
 
-        if (dirDialog.exec() != QDialog::Accepted) {
-            return 1;
-        }
-
-        const auto urls = dirDialog.selectedUrls();
-        if (urls.empty()) {
-            return 2;
-        }
-
-        results.insert(QStringLiteral("uris"), fuseRedirect(urls));
-        results.insert(QStringLiteral("writable"), true);
-
-        return 0;
+        delayReply(message, dirDialog, this, [dirDialog](int result) -> QVariantList {
+            dirDialog->deleteLater();
+            if (result != QDialog::Accepted) {
+                return {PortalResponse::Cancelled, QVariantMap{}};
+            }
+            const auto urls = dirDialog->selectedUrls();
+            if (urls.empty()) {
+                return {PortalResponse::OtherError, QVariantMap{}};
+            }
+            return {PortalResponse::Success, QVariantMap{{QStringLiteral("uris"), fuseRedirect(urls)}, {QStringLiteral("writable"), true}}};
+        });
+        dirDialog->open();
+        return;
     }
 
     // for handling of options - choices
-    QScopedPointer<QWidget> optionsWidget;
+    QWidget *optionsWidget = nullptr;
     // to store IDs for choices along with corresponding comboboxes/checkboxes
     QMap<QString, QCheckBox *> checkboxes;
     QMap<QString, QComboBox *> comboboxes;
 
+    auto fileDialog = new FileDialog();
+    Utils::setParentWindow(fileDialog, parent_window);
+    Request::makeClosableDialogRequest(handle, fileDialog);
+    fileDialog->setWindowModality(modalDialog ? Qt::WindowModal : Qt::NonModal);
+
     if (options.contains(QStringLiteral("choices"))) {
         OptionList optionList = qdbus_cast<OptionList>(options.value(QStringLiteral("choices")));
-        optionsWidget.reset(CreateChoiceControls(optionList, checkboxes, comboboxes));
+        optionsWidget = CreateChoiceControls(optionList, checkboxes, comboboxes);
     }
 
-    QScopedPointer<FileDialog, QScopedPointerDeleteLater> fileDialog(new FileDialog());
-    Utils::setParentWindow(fileDialog.data(), parent_window);
-    Request::makeClosableDialogRequest(handle, fileDialog.get());
     fileDialog->setWindowTitle(title);
-    fileDialog->setModal(modalDialog);
     KFile::Mode mode = directory ? KFile::Mode::Directory : multipleFiles ? KFile::Mode::Files : KFile::Mode::File;
     fileDialog->m_fileWidget->setMode(mode | KFile::Mode::ExistingOnly);
     if (!isKIOFuseAvailable()) {
@@ -485,39 +455,40 @@ uint FileChooserPortal::OpenFile(const QDBusObjectPath &handle,
     fileDialog->m_fileWidget->setFilters(filters, currentFilter);
 
     if (optionsWidget) {
-        fileDialog->m_fileWidget->setCustomWidget({}, optionsWidget.get());
+        fileDialog->m_fileWidget->setCustomWidget({}, optionsWidget);
+        optionsWidget->setParent(fileDialog);
     }
 
-    if (fileDialog->exec() == QDialog::Accepted) {
-        const auto urls = fileDialog->m_fileWidget->selectedUrls();
-        if (urls.isEmpty()) {
-            qCDebug(XdgDesktopPortalKdeFileChooser) << "Failed to open file: no local file selected";
-            return 2;
+    delayReply(message, fileDialog, this, [fileDialog, optionsWidget, checkboxes, comboboxes](int dialogResult) -> QVariantList {
+        fileDialog->deleteLater();
+        QVariantMap results;
+        if (dialogResult == QDialog::Accepted) {
+            const auto urls = fileDialog->m_fileWidget->selectedUrls();
+            if (urls.isEmpty()) {
+                return {PortalResponse::Cancelled, results};
+            }
+            results.insert(QStringLiteral("uris"), fuseRedirect(urls));
+            results.insert(QStringLiteral("writable"), true);
+            if (optionsWidget) {
+                QVariant choices = EvaluateSelectedChoices(checkboxes, comboboxes);
+                results.insert(QStringLiteral("choices"), choices);
+            }
+            FilterList currentFilter = fileFilterToFilterList(fileDialog->m_fileWidget->currentFilter());
+            results.insert(QStringLiteral("current_filter"), QVariant::fromValue<FilterList>(currentFilter));
         }
-
-        results.insert(QStringLiteral("uris"), fuseRedirect(urls));
-        results.insert(QStringLiteral("writable"), true);
-
-        if (optionsWidget) {
-            QVariant choices = EvaluateSelectedChoices(checkboxes, comboboxes);
-            results.insert(QStringLiteral("choices"), choices);
-        }
-
-        FilterList currentFilter = fileFilterToFilterList(fileDialog->m_fileWidget->currentFilter());
-        results.insert(QStringLiteral("current_filter"), QVariant::fromValue<FilterList>(currentFilter));
-
-        return 0;
-    }
-
-    return 1;
+        return {PortalResponse::fromDialogCode(static_cast<QDialog::DialogCode>(dialogResult)), results};
+    });
+    fileDialog->open();
 }
 
-uint FileChooserPortal::SaveFile(const QDBusObjectPath &handle,
+void FileChooserPortal::SaveFile(const QDBusObjectPath &handle,
                                  const QString &app_id,
                                  const QString &parent_window,
                                  const QString &title,
                                  const QVariantMap &options,
-                                 QVariantMap &results)
+                                 const QDBusMessage &message,
+                                 uint &replyResponse,
+                                 QVariantMap &replyResults)
 {
     Q_UNUSED(app_id);
 
@@ -553,7 +524,9 @@ uint FileChooserPortal::SaveFile(const QDBusObjectPath &handle,
     const auto [filters, currentFilter] = ExtractFilters(options);
 
     if (isMobile()) {
-        if (!m_mobileFileDialog) {
+        if (m_mobileFileDialog) {
+            m_mobileFileDialog->reset(true);
+        } else {
             qCDebug(XdgDesktopPortalKdeFileChooser) << "Creating file dialog";
             m_mobileFileDialog = new MobileFileDialog(this);
         }
@@ -586,29 +559,30 @@ uint FileChooserPortal::SaveFile(const QDBusObjectPath &handle,
             m_mobileFileDialog->setMimeTypeFilters(filters.first().mimePatterns());
         }
 
-        uint retCode = m_mobileFileDialog->exec();
+        replyResponse = m_mobileFileDialog->exec();
 
-        results.insert(QStringLiteral("uris"), fuseRedirect(m_mobileFileDialog->results()));
+        replyResults.insert(QStringLiteral("uris"), fuseRedirect(m_mobileFileDialog->results()));
 
-        return retCode;
+        return;
     }
 
     // for handling of options - choices
-    QScopedPointer<QWidget> optionsWidget;
+    QWidget *optionsWidget = nullptr;
     // to store IDs for choices along with corresponding comboboxes/checkboxes
     QMap<QString, QCheckBox *> checkboxes;
     QMap<QString, QComboBox *> comboboxes;
 
+    auto fileDialog = new FileDialog();
+    Utils::setParentWindow(fileDialog, parent_window);
+    Request::makeClosableDialogRequest(handle, fileDialog);
+    fileDialog->setWindowModality(modalDialog ? Qt::WindowModal : Qt::NonModal);
+
     if (options.contains(QStringLiteral("choices"))) {
         OptionList optionList = qdbus_cast<OptionList>(options.value(QStringLiteral("choices")));
-        optionsWidget.reset(CreateChoiceControls(optionList, checkboxes, comboboxes));
+        optionsWidget = CreateChoiceControls(optionList, checkboxes, comboboxes);
     }
 
-    QScopedPointer<FileDialog, QScopedPointerDeleteLater> fileDialog(new FileDialog());
-    Utils::setParentWindow(fileDialog.data(), parent_window);
-    Request::makeClosableDialogRequest(handle, fileDialog.get());
     fileDialog->setWindowTitle(title);
-    fileDialog->setModal(modalDialog);
     fileDialog->m_fileWidget->setOperationMode(KFileWidget::Saving);
     fileDialog->m_fileWidget->setConfirmOverwrite(true);
 
@@ -648,30 +622,29 @@ uint FileChooserPortal::SaveFile(const QDBusObjectPath &handle,
     fileDialog->m_fileWidget->setFilters(filters, currentFilter);
 
     if (optionsWidget) {
-        fileDialog->m_fileWidget->setCustomWidget(optionsWidget.get());
+        fileDialog->m_fileWidget->setCustomWidget(optionsWidget);
+        optionsWidget->setParent(fileDialog);
     }
 
-    if (fileDialog->exec() == QDialog::Accepted) {
-        const auto urls = fileDialog->m_fileWidget->selectedUrls();
-        results.insert(QStringLiteral("uris"), fuseRedirect(urls));
-
-        if (optionsWidget) {
-            QVariant choices = EvaluateSelectedChoices(checkboxes, comboboxes);
-            results.insert(QStringLiteral("choices"), choices);
+    delayReply(message, fileDialog, this, [fileDialog, optionsWidget, checkboxes, comboboxes](int dialogResult) {
+        QVariantMap results;
+        fileDialog->deleteLater();
+        if (dialogResult == QDialog::Accepted) {
+            const auto urls = fileDialog->m_fileWidget->selectedUrls();
+            results.insert(QStringLiteral("uris"), fuseRedirect(urls));
+            if (optionsWidget) {
+                QVariant choices = EvaluateSelectedChoices(checkboxes, comboboxes);
+                results.insert(QStringLiteral("choices"), choices);
+            }
+            FilterList currentFilter = fileFilterToFilterList(fileDialog->m_fileWidget->currentFilter());
+            results.insert(QStringLiteral("current_filter"), QVariant::fromValue<FilterList>(currentFilter));
         }
-
-        FilterList currentFilter = fileFilterToFilterList(fileDialog->m_fileWidget->currentFilter());
-        results.insert(QStringLiteral("current_filter"), QVariant::fromValue<FilterList>(currentFilter));
-
-        return 0;
-    }
-
-    return 1;
+        return QVariantList{PortalResponse::fromDialogCode(static_cast<QDialog::DialogCode>(dialogResult)), results};
+    });
+    fileDialog->open();
 }
 
-QWidget *FileChooserPortal::CreateChoiceControls(const FileChooserPortal::OptionList &optionList,
-                                                 QMap<QString, QCheckBox *> &checkboxes,
-                                                 QMap<QString, QComboBox *> &comboboxes)
+QWidget *FileChooserPortal::CreateChoiceControls(const OptionList &optionList, QMap<QString, QCheckBox *> &checkboxes, QMap<QString, QComboBox *> &comboboxes)
 {
     if (optionList.empty()) {
         return nullptr;
