@@ -112,12 +112,12 @@ uint RemoteDesktopPortal::CreateSession(const QDBusObjectPath &handle,
     Session *session = Session::createSession(this, Session::RemoteDesktop, app_id, session_handle.path());
 
     if (!session) {
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     if (!WaylandIntegration::isStreamingAvailable()) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "zkde_screencast_unstable_v1 does not seem to be available";
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     connect(session, &Session::closed, [session] {
@@ -133,7 +133,7 @@ uint RemoteDesktopPortal::CreateSession(const QDBusObjectPath &handle,
         }
     });
 
-    return 0;
+    return PortalResponse::Success;
 }
 
 uint RemoteDesktopPortal::SelectDevices(const QDBusObjectPath &handle,
@@ -152,91 +152,26 @@ uint RemoteDesktopPortal::SelectDevices(const QDBusObjectPath &handle,
     const auto types = static_cast<RemoteDesktopPortal::DeviceTypes>(options.value(QStringLiteral("types")).toUInt());
     if (types == None) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "There are no devices to remotely control";
-        return 2;
+        return PortalResponse::OtherError;
     }
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to select sources on non-existing session " << session_handle.path();
-        return 2;
+        return PortalResponse::OtherError;
     }
 
     session->setDeviceTypes(types);
     session->setPersistMode(ScreenCastPortal::PersistMode(options.value(QStringLiteral("persist_mode")).toUInt()));
     session->setRestoreData(options.value(QStringLiteral("restore_data")));
 
-    return 0;
+    return PortalResponse::Success;
 }
 
-uint RemoteDesktopPortal::Start(const QDBusObjectPath &handle,
-                                const QDBusObjectPath &session_handle,
-                                const QString &app_id,
-                                const QString &parent_window,
-                                const QVariantMap &options,
-                                QVariantMap &results)
+std::pair<PortalResponse::Response, QVariantMap> continueStart(RemoteDesktopSession *session, ScreenCastPortal::PersistMode persist)
 {
-    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Start called with parameters:";
-    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    handle: " << handle.path();
-    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    session_handle: " << session_handle.path();
-    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    app_id: " << app_id;
-    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    parent_window: " << parent_window;
-    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    options: " << options;
-
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
-
-    if (!session) {
-        qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call start on non-existing session " << session_handle.path();
-        return 2;
-    }
-
-    if (QGuiApplication::screens().isEmpty()) {
-        qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Failed to show dialog as there is no screen to select";
-        return 2;
-    }
-
-    const ScreenCastPortal::PersistMode persist = session->persistMode();
-    QList<Output> selectedOutputs;
-
-    bool restored = false;
-
-    if (persist != ScreenCastPortal::NoPersist && session->restoreData().isValid()) {
-        const RestoreData restoreData = qdbus_cast<RestoreData>(session->restoreData().value<QDBusArgument>());
-        if (restoreData.session == QLatin1String("KDE") && restoreData.version == RestoreData::currentRestoreDataVersion()) {
-            // check we asked for the same key content both times; if not, don't restore
-            // some settings (like ScreenCast multipleSources or cursorMode) don't involve user prompts so use whatever was explicitly
-            // requested this time
-            const RemoteDesktopPortal::DeviceTypes devices = static_cast<RemoteDesktopPortal::DeviceTypes>(restoreData.payload["devices"].toUInt());
-            if (session->deviceTypes() != devices) {
-                qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Not restoring session as requested devices don't match";
-            } else if (session->screenSharingEnabled() != restoreData.payload["screenShareEnabled"].toBool()) {
-                qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Not restoring session as requested screen sharing doesn't match";
-            } else {
-                restored = true;
-            }
-        }
-    }
-
-    if (restored) {
-        auto notification = new KNotification(QStringLiteral("remotedesktopstarted"), KNotification::CloseOnTimeout);
-        notification->setTitle(i18nc("title of notification about input systems taken over", "Remote control session started"));
-        notification->setText(RemoteDesktopDialog::buildDescription(app_id, session->deviceTypes(), session->screenSharingEnabled()));
-        notification->setIconName(QStringLiteral("krfb"));
-        notification->sendEvent();
-    } else {
-        if (!isAppMegaAuthorized(app_id)) { // authorize right away
-            QScopedPointer<RemoteDesktopDialog, QScopedPointerDeleteLater> remoteDesktopDialog(
-                new RemoteDesktopDialog(app_id, session->deviceTypes(), session->screenSharingEnabled(), session->persistMode()));
-            Utils::setParentWindow(remoteDesktopDialog->windowHandle(), parent_window);
-            Request::makeClosableDialogRequest(handle, remoteDesktopDialog.get());
-            connect(session, &Session::closed, remoteDesktopDialog.data(), &RemoteDesktopDialog::reject);
-
-            if (!remoteDesktopDialog->exec()) {
-                return 1;
-            }
-        }
-    }
-
+    QVariantMap results;
     if (session->screenSharingEnabled()) {
         WaylandIntegration::Streams streams;
         const auto screens = qGuiApp->screens();
@@ -244,7 +179,7 @@ uint RemoteDesktopPortal::Start(const QDBusObjectPath &handle,
             for (const auto &screen : screens) {
                 auto stream = WaylandIntegration::startStreamingOutput(screen, Screencasting::Metadata);
                 if (!stream.isValid()) {
-                    return 2;
+                    return {PortalResponse::OtherError, {}};
                 }
                 streams << stream;
             }
@@ -263,17 +198,97 @@ uint RemoteDesktopPortal::Start(const QDBusObjectPath &handle,
     results.insert(QStringLiteral("devices"), QVariant::fromValue<uint>(session->deviceTypes()));
     results.insert(QStringLiteral("clipboard_enabled"), false);
     if (session->persistMode() != ScreenCastPortal::NoPersist) {
-        results.insert("persist_mode", quint32(persist));
+        results.insert(u"persist_mode"_s, quint32(persist));
         if (persist != ScreenCastPortal::NoPersist) {
-            const RestoreData restoreData = {
-                "KDE",
-                RestoreData::currentRestoreDataVersion(),
-                QVariantMap{{"screenShareEnabled", session->screenSharingEnabled()}, {"devices", static_cast<quint32>(session->deviceTypes())}}};
-            results.insert("restore_data", QVariant::fromValue<RestoreData>(restoreData));
+            const RestoreData restoreData = {u"KDE"_s,
+                                             RestoreData::currentRestoreDataVersion(),
+                                             QVariantMap{{u"screenShareEnabled"_s, session->screenSharingEnabled()},
+                                                         {u"devices"_s, static_cast<quint32>(session->deviceTypes())},
+                                                         {u"clipboardEnabled"_s, session->clipboardEnabled()}}};
+            results.insert(u"restore_data"_s, QVariant::fromValue<RestoreData>(restoreData));
+        }
+    }
+    return {PortalResponse::Success, results};
+}
+
+void RemoteDesktopPortal::Start(const QDBusObjectPath &handle,
+                                const QDBusObjectPath &session_handle,
+                                const QString &app_id,
+                                const QString &parent_window,
+                                const QVariantMap &options,
+                                const QDBusMessage &message,
+                                uint &replyResponse,
+                                QVariantMap &replyResults)
+{
+    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Start called with parameters:";
+    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    handle: " << handle.path();
+    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    session_handle: " << session_handle.path();
+    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    app_id: " << app_id;
+    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    parent_window: " << parent_window;
+    qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    options: " << options;
+
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
+
+    if (!session) {
+        qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call start on non-existing session " << session_handle.path();
+        replyResponse = PortalResponse::OtherError;
+        return;
+    }
+
+    if (QGuiApplication::screens().isEmpty()) {
+        qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Failed to show dialog as there is no screen to select";
+        replyResponse = PortalResponse::OtherError;
+        return;
+    }
+
+    const ScreenCastPortal::PersistMode persist = session->persistMode();
+    QList<Output> selectedOutputs;
+
+    bool restored = false;
+
+    if (persist != ScreenCastPortal::NoPersist && session->restoreData().isValid()) {
+        const RestoreData restoreData = qdbus_cast<RestoreData>(session->restoreData().value<QDBusArgument>());
+        if (restoreData.session == QLatin1String("KDE") && restoreData.version == RestoreData::currentRestoreDataVersion()) {
+            // check we asked for the same key content both times; if not, don't restore
+            // some settings (like ScreenCast multipleSources or cursorMode) don't involve user prompts so use whatever was explicitly
+            // requested this time
+            const RemoteDesktopPortal::DeviceTypes devices = static_cast<RemoteDesktopPortal::DeviceTypes>(restoreData.payload[u"devices"_s].toUInt());
+            if (session->deviceTypes() != devices) {
+                qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Not restoring session as requested devices don't match";
+            } else if (session->screenSharingEnabled() != restoreData.payload[u"screenShareEnabled"_s].toBool()) {
+                qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Not restoring session as requested screen sharing doesn't match";
+            } else if (session->clipboardEnabled() != restoreData.payload[u"clipboardEnabled"_s].toBool()) {
+                qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "Not restoring session as clipboard enabled doesn't match";
+            } else {
+                restored = true;
+            }
         }
     }
 
-    return 0;
+    if (restored) {
+        auto notification = new KNotification(QStringLiteral("remotedesktopstarted"), KNotification::CloseOnTimeout);
+        notification->setTitle(i18nc("title of notification about input systems taken over", "Remote control session started"));
+        notification->setText(RemoteDesktopDialog::buildDescription(app_id, session->deviceTypes(), session->screenSharingEnabled()));
+        notification->setIconName(QStringLiteral("krfb"));
+        notification->sendEvent();
+    } else {
+        if (!isAppMegaAuthorized(app_id)) { // authorize right away
+            auto remoteDesktopDialog = new RemoteDesktopDialog(app_id, session->deviceTypes(), session->screenSharingEnabled(), session->persistMode());
+            Utils::setParentWindow(remoteDesktopDialog->windowHandle(), parent_window);
+            Request::makeClosableDialogRequestWithSession(handle, remoteDesktopDialog, session);
+            delayReply(message, remoteDesktopDialog, this, [session, persist](DialogResult dialogResult) {
+                auto response = PortalResponse::fromDialogResult(dialogResult);
+                QVariantMap results;
+                if (dialogResult == DialogResult::Accepted) {
+                    std::tie(response, results) = continueStart(session, persist);
+                }
+                return QVariantList{response, results};
+            });
+            return;
+        }
+    }
+
+    std::tie(replyResponse, replyResults) = continueStart(session, persist);
 }
 
 void RemoteDesktopPortal::NotifyPointerMotion(const QDBusObjectPath &session_handle, const QVariantMap &options, double dx, double dy)
@@ -284,7 +299,7 @@ void RemoteDesktopPortal::NotifyPointerMotion(const QDBusObjectPath &session_han
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    dx: " << dx;
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    dy: " << dy;
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerMotion on non-existing session " << session_handle.path();
@@ -303,7 +318,7 @@ void RemoteDesktopPortal::NotifyPointerMotionAbsolute(const QDBusObjectPath &ses
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    x: " << x;
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    y: " << y;
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerMotionAbsolute on non-existing session " << session_handle.path();
@@ -321,7 +336,7 @@ void RemoteDesktopPortal::NotifyPointerButton(const QDBusObjectPath &session_han
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    button: " << button;
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    state: " << state;
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerButton on non-existing session " << session_handle.path();
@@ -343,7 +358,7 @@ void RemoteDesktopPortal::NotifyPointerAxis(const QDBusObjectPath &session_handl
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    dx: " << dx;
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    dy: " << dy;
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyKeyboardKeysym on non-existing session " << session_handle.path();
@@ -361,7 +376,7 @@ void RemoteDesktopPortal::NotifyPointerAxisDiscrete(const QDBusObjectPath &sessi
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    axis: " << axis;
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    steps: " << steps;
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerAxisDiscrete on non-existing session " << session_handle.path();
@@ -375,7 +390,7 @@ void RemoteDesktopPortal::NotifyKeyboardKeysym(const QDBusObjectPath &session_ha
 {
     Q_UNUSED(options)
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyKeyboardKeysym on non-existing session " << session_handle.path();
@@ -393,7 +408,7 @@ void RemoteDesktopPortal::NotifyKeyboardKeycode(const QDBusObjectPath &session_h
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    keycode: " << keycode;
     qCDebug(XdgDesktopPortalKdeRemoteDesktop) << "    state: " << state;
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
 
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyKeyboardKeycode on non-existing session " << session_handle.path();
@@ -408,7 +423,7 @@ void RemoteDesktopPortal::NotifyTouchDown(const QDBusObjectPath &session_handle,
     Q_UNUSED(options)
     Q_UNUSED(stream)
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerAxisDiscrete on non-existing session " << session_handle.path();
         return;
@@ -421,7 +436,7 @@ void RemoteDesktopPortal::NotifyTouchMotion(const QDBusObjectPath &session_handl
     Q_UNUSED(options)
     Q_UNUSED(stream)
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerAxisDiscrete on non-existing session " << session_handle.path();
         return;
@@ -433,7 +448,7 @@ void RemoteDesktopPortal::NotifyTouchUp(const QDBusObjectPath &session_handle, c
 {
     Q_UNUSED(options)
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call NotifyPointerAxisDiscrete on non-existing session " << session_handle.path();
         return;
@@ -448,7 +463,7 @@ RemoteDesktopPortal::ConnectToEIS(const QDBusObjectPath &session_handle, const Q
     Q_UNUSED(options)
     Q_UNUSED(app_id)
 
-    RemoteDesktopSession *session = qobject_cast<RemoteDesktopSession *>(Session::getSession(session_handle.path()));
+    RemoteDesktopSession *session = Session::getSession<RemoteDesktopSession>(session_handle.path());
     if (!session) {
         qCWarning(XdgDesktopPortalKdeRemoteDesktop) << "Tried to call ConnectToEis on non-existing session " << session_handle.path();
         return QDBusUnixFileDescriptor();
