@@ -21,26 +21,29 @@
 #include <QStandardPaths>
 
 #include <KLocalizedString>
+#include <KService>
 
-UserInfoDialog::UserInfoDialog(const QString &reason, QObject *parent)
+using namespace Qt::StringLiterals;
+
+UserInfoDialog::UserInfoDialog(const QString &reason, const QString &app_id, QObject *parent)
     : QuickDialog(parent)
 {
     QString ifacePath = QStringLiteral("/org/freedesktop/Accounts/User%1").arg(getuid());
     m_userInterface = new OrgFreedesktopAccountsUserInterface(QStringLiteral("org.freedesktop.Accounts"), ifacePath, QDBusConnection::systemBus(), this);
 
-    QString image = QFileInfo::exists(m_userInterface->iconFile()) ? m_userInterface->iconFile() : QString();
-    QVariantMap props = {{"title", i18n("Share Information")},
-                         {"subtitle", i18n("Share your personal information with the requesting application?")},
-                         {"reason", reason},
-                         {"userName", m_userInterface->userName()},
-                         {"realName", m_userInterface->realName()}};
+    const KService::Ptr app = KService::serviceByDesktopName(app_id);
+    const QString appName = app ? app->name() : app_id;
+
+    QVariantMap props = {{u"title"_s, i18n("Share User Information with %1", appName)},
+                         {u"subtitle"_s, reason.isEmpty() ? i18n("%1 wants access to your user info.", appName) : reason},
+                         {u"iconName"_s, QStringLiteral("user-identity")},
+                         {u"details"_s, i18n("This app will be given access to your avatar and name.")},
+                         {u"name"_s, m_userInterface->realName().isEmpty() ? m_userInterface->userName() : m_userInterface->realName()}};
 
     if (QFileInfo::exists(m_userInterface->iconFile())) {
-        props.insert(QStringLiteral("iconName"), m_userInterface->iconFile());
-    } else {
-        props.insert(QStringLiteral("iconName"), QStringLiteral("user-identity"));
+        props.insert(QStringLiteral("avatar"), image());
     }
-    create("qrc:/UserInfoDialog.qml", props);
+    create(u"qrc:/UserInfoDialog.qml"_s, props);
 }
 
 UserInfoDialog::~UserInfoDialog()
@@ -54,6 +57,9 @@ QString UserInfoDialog::id() const
 
 QString UserInfoDialog::image() const
 {
+    if (!QFileInfo::exists(m_userInterface->iconFile())) {
+        return {};
+    }
     return QUrl::fromLocalFile(m_userInterface->iconFile()).toString();
 }
 
